var util = require('util');
var events = require('events');
var xmpp = require('node-xmpp');
var bind = require('underscore').bind;

var Bot;

module.exports.Bot = (function() {
  // Private stuff

  // Called whenever an XMPP error occurs
  var onError = function(error) {
    if (error instanceof xmpp.Element) {
      var errorMessage = 'Unknown error';

      // Check of see-other-host error
      if (error.getChild('see-other-host')) {
        errorMessage = 'This account is signing in somewhere else. Maybe HipChat web version?';
      } else {
        // Handle XMPP text errors (when available)
        var textNode = error.getChild('text');

        if (textNode) {
          errorMessage = textNode.getText();
        }
      }

      this.emit('error', errorMessage, error);
    } else {
      this.emit('error', error);
    }

    this.disconnect();
  };

  // Called when debug is enabled
  var enableDebug = function() {
    var self = this;

    self.jabber.on('data', function(buffer) {
      console.log('  IN > ' + buffer.toString());
    });

    var origSend = this.jabber.send;
    self.jabber.send = function(stanza) {
      console.log(' OUT > ' + stanza);
      return origSend.call(self.jabber, stanza);
    };
  };

  // Called whenever XMPP connection is made
  var onOnline = function() {
    var self = this;

    // Send online presence
    self.setAvailability('chat');

    // Server will disconnect us after 150s of inactivity
    self.keepalive = setInterval(function() {
      self.jabber.send(' ');
      self.emit('ping');
    }, 30000);

    self.emit('connect');
  };

  // Called whenever there is an incoming XMPP stanza
  var onStanza = function(stanza) {
    this.emit('data', stanza);

    if (stanza.is('message') && stanza.attrs.type == 'groupchat') {
      var body = stanza.getChild('body');
      if (!body) return;

      if (stanza.getChild('delay')) return;

      var attrFrom = stanza.attrs.from;
      var offset = attrFrom.indexOf('/');

      var channel = attrFrom.substring(0, offset);
      var from = attrFrom.substring(offset + 1);

      // Do not trigger when message is from self.
      if (from == this.name) return;

      this.emit('message', channel, from, body.getText());
    } else if (stanza.is('message') && stanza.attrs.type == 'chat') {
      var body = stanza.getChild('body');
      if (!body) return;

      var attrFrom = stanza.attrs.from;
      var offset = attrFrom.indexOf('/');

      var jid = attrFrom.substring(0, offset);

      this.emit('privateMessage', jid, body.getText());
    }
  };

  // Public API

  // Constructor
  // `options`: Object
  //    - `jid`: Bot's jid
  //    - `password`: Bot's HipChat password
  //    - `name`: Bot's full name (must match HipChat's data)
  //    - `debug`: Set to `true` to show network debug in console
  function Bot(options) {
    events.EventEmitter.call(this);
    this.once('connect', function() { }); // listener bug
    this.setMaxListeners(0);

    this.jabber = null;
    this.keepalive = null;
    this.plugins = {};

    options = options || {};
    this.jid = options.jid;
    this.password = options.password;
    this.debug = options.debug;
    this.name = options.name;
  };

  util.inherits(Bot, events.EventEmitter);

  // `connect()`
  // Connects to HipChat.
  Bot.prototype.connect = function() {
    // Create XMPP connection.
    this.jabber = new xmpp.Client({
      jid: this.jid,
      password: this.password
    });

    // If `options.debug` is set to `true`, all incoming
    // and outgoing XMPP packets will be printed to the console.
    if (this.debug === true) {
      enableDebug.call(this);
    }

    // Bind to XMPP events.
    this.jabber.on('error', bind(onError, this));
    this.jabber.on('online', bind(onOnline, this));
    this.jabber.on('stanza', bind(onStanza, this));
  };

  // `setAvailability(availability, status)`
  // Update availability status.
  // - `availability`: Jabber availability codes
  //    - `away`
  //    - `chat` (Free for chat)
  //    - `dnd` (Do not disturb)
  // - `status`: Status message to display
  Bot.prototype.setAvailability = function(availability, status) {
    var packet = new xmpp.Element('presence', { type: 'available' });
    packet.c('show').t(availability);

    if (status) {
      packet.c('status').t(status);
    }

    this.jabber.send(packet);
  };

  // `join(roomJid)`
  // Join the specified room.
  // - `roomJid`: Target room, in the form of `????_????@conf.hipchat.com`
  Bot.prototype.join = function(roomJid) {
    var packet = new xmpp.Element('presence', { to: roomJid + '/' + this.name });
    packet.c('x', { xmlns: 'http://jabber.org/protocol/muc' });
    this.jabber.send(packet);
  };

  // `part(roomJid)`
  // Part the specified room.
  // - `roomJid`: Target room, in the form of `????_????@conf.hipchat.com`
  Bot.prototype.part = function(roomJid) {
    var packet = new xmpp.Element('presence', { type: 'unavailable', to: roomJid + '/' + this.name });
    packet.c('x', { xmlns: 'http://jabber.org/protocol/muc' });
    packet.c('status').t('hc-leave');
    this.jabber.send(packet);
  };

  // `message(roomJid, message)`
  // Send a message to a room or a user.
  // - `targetJid`: Target
  //    - Message to a room: `????_????@conf.hipchat.com`
  //    - Private message to a user: `????_????@chat.hipchat.com`
  // - `message`: Message to be sent to the room
  Bot.prototype.message = function(targetJid, message) {
    var packet;

    if (targetJid.match(/^(.*)@conf.hipchat.com$/)) {
      packet = new xmpp.Element('message', {
        to: targetJid + '/' + this.name,
        type: 'groupchat'
      });
    } else {
      packet = new xmpp.Element('message', {
        to: targetJid,
        type: 'chat',
        from: this.jid
      });
      packet.c('inactive', { xmlns: 'http://jabber/protocol/chatstates' });
    }

    packet.c('body').t(message);
    this.jabber.send(packet);
  };

  // `disconnect()`
  // Disconnect the bot from HipChat.
  Bot.prototype.disconnect = function() {
    if (this.keepalive) {
      clearInterval(this.keepalive);
      this.keepalive = null;
    }
    this.jabber.end();
    this.emit('disconnect');
  };

  Bot.prototype.loadPlugin = function(identifier, plugin, options) {
    if (typeof(plugin) !== 'object') {
      throw new Error('plugin argument must be an object');
    }

    if (typeof(plugin.load) !== 'function') {
      throw new Error('plugin object must have a load function');
    }

    this.plugins[identifier] = plugin;
    this.plugins[identifier].load(this, options);

    return true;
  };

  // Events API

  // `onConnect(callback)`
  // Emitted whenever the bot connects to the server.
  // - `callback`: Function to be triggered: `function ()`
  Bot.prototype.onConnect = function(callback) {
    this.on('connect', callback);
  };

  // `onMessage(condition, callback)`
  // `onMessage(callback)`
  // Emitted whenever a message is sent to a channel the bot is in.
  // - `condition`: String or RegExp the message must match.
  // - `callback`: Function to be triggered: `function (roomJid, from, message, matches)`
  //   - `roomJid`: Jabber Id of the room in which the message occured.
  //   - `from`: The name of the person who said the message.
  //   - `message`: The message
  //   - `matches`: The matches returned by the condition when it is a RegExp
  Bot.prototype.onMessage = function(condition, callback) {
    if (arguments.length === 1) {
      return this.on('message', condition);
    }

    this.on('message', function(channel, from, message) {
      if (typeof condition === 'string' && message === condition) {
        callback.apply(this, arguments);
      } else if (condition instanceof RegExp) {
        var matches = message.match(condition);

        if (!matches) {
          return;
        }

        var args = Array.prototype.slice.apply(arguments);
        args.push(matches);

        callback.apply(this, args);
      }
    });
  };

  // `onPrivateMessage(condition, callback)`
  // `onPrivateMessage(callback)`
  // Emitted whenever a message is sent privately to the bot.
  // - `condition`: String or RegExp the message must match.
  // - `callback`: Function to be triggered: `function (fromJid, message)`
  Bot.prototype.onPrivateMessage = function(condition, callback) {
    if (arguments.length === 1) {
      return this.on('privateMessage', condition);
    }

    this.on('privateMessage', function(from, message) {
      if (typeof condition === 'string' && message === condition) {
        callback.apply(this, arguments);
      } else if (condition instanceof RegExp) {
        var matches = message.match(condition);

        if (!matches) {
          return;
        }

        var args = Array.prototype.slice.apply(arguments);
        args.push(matches);

        callback.apply(this, args);
      }
    });
  };

  // `onPing(callback)`
  // Emitted whenever the bot pings the server (roughly every 30 seconds).
  // - `callback`: Function to be triggered: `function ()`
  Bot.prototype.onPing = function(callback) {
    this.on('ping', callback);
  };

  // `onError(callback)`
  // Emitted whenever an error occurs. `disconnect` will be emitted afterwards.
  // - `callback`: Function to be triggered: `function(message, stanza)`
  //   - `message`: string representation of the error.
  //   - `stanza`: instance of `xmpp.Element` (when available)
  Bot.prototype.onError = function(callback) {
    this.on('error', callback);
  };

  // `onDisconnect(callback)`
  // Emitted whenever the bot disconnects from the server.
  // - `callback`: Function to be triggered: `function ()`
  Bot.prototype.onDisconnect = function(callback) {
    this.on('disconnect', callback);
  };

  return Bot;
}());
